#    This is a part of the "Prototype Tagger", an unsupervised POS tagger developed
#    in the Hebrew University by Omri Abend, Roi Reichart and Ari Rappoport.
#    Copyright (C) 2010  Omri Abend, Roi Reichart and Ari Rappoport
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    If you have any questions, feel free to contact us in:
#    omria01@cs.huji.ac.il, roiri@cs.huji.ac.il
#
#    Please read the README file for more details.
#
#!/usr/local/bin/python
import random
from utils import *
from matrix import *
from sequencer import *
from math import sqrt

USE_RANDOM_REPRESENTATIVE = False
USE_BIGGER_CLUSTER_REPRESENTATIVE = False
USE_LAST_MODIFIED_REPRESENTATIVE = True

class average_cos:
    
    def __init__(self, thresh, n):
        self._matrix = matrix(n)
        self._thresh = thresh
        self._sequencer = sequencer()
        self._initial_cluster_lists = None
        self._sizes_of_clusters = [0 for index in range(self._matrix.size())]
        self._freq_dict = None
        self._linking_history = []  # Keeps a list of the lately merged lists. The last is the most recent.
                                    # Uses this to select a representative.
        
    def load_matrix_and_sequencer(self, inp_matrix, inp_sequencer):
        self._matrix = inp_matrix
        self._sequencer = inp_sequencer
        self._sizes_of_clusters = [1 for index in range(inp_matrix._n)]
    
    def load_freq_dict(self, freq_dict):
        self._freq_dict = freq_dict
    
    def load_matrix_from_dict(self, vect_dict):
        for key1 in vect_dict.keys():
            for key2 in vect_dict.keys():
                if key1 == key2 or key1 > key2:
                    continue
                self._matrix.put(self._sequencer.get(key1),
                                 self._sequencer.get(key2),
                                 cos(vect_dict[key1], vect_dict[key2]))
        self._sizes_of_clusters = [1 for index in range(self._matrix.size())]
    
    def load_matrix_from_dict_with_initial_clustering(self, vect_dict, initial_cluster_lists):
        self._initial_cluster_lists = initial_cluster_lists
        for cluster_name1 in initial_cluster_lists.keys():
            for cluster_name2 in initial_cluster_lists.keys():
                if cluster_name1 == cluster_name2:
                    continue
                self._matrix.put(self._sequencer.get(cluster_name1), self._sequencer.get(cluster_name2), 
                                 cos_clusters(initial_cluster_lists[cluster_name1],
                                              initial_cluster_lists[cluster_name2], vect_dict))
        for key in initial_cluster_lists.keys():
            self._sizes_of_clusters[self._sequencer.get(key)] = len(initial_cluster_lists[key])
    
    def load_matrix(self, matrix):
        self._matrix = matrix
    
    def avg_link_clustering(self):
        index_to_cluster = dict([(k,k) for k in range(self._matrix.size())])
        cluster_lists = dict([(k,[k]) for k in range(self._matrix.size())])
        self._linking_history = []
        
        number_of_iterations = 0
        while 1:
            if len(cluster_lists.keys()) <= 1:
                break
            max_key, max_val = self._matrix.max_matrix()
            #print self._sequencer.inv_get(max_key[0]), self._sequencer.inv_get(max_key[1]), max_val 
            if max_val < self._thresh:
                break
            
            len1 = self._sizes_of_clusters[max_key[0]]
            len2 = self._sizes_of_clusters[max_key[1]]
            new_cluster_name = self.get_representative(max_key[0], max_key[1])

            if new_cluster_name == max_key[0]:
                self._matrix.merge_rows_columns(max_key[0], max_key[1], 
                                            1.0*len1/(len1+len2), 1.0*len2/(len1+len2))
            else:
                self._matrix.merge_rows_columns(max_key[1], max_key[0], 
                                            1.0*len2/(len1+len2), 1.0*len1/(len1+len2))
            
            new_cluster = cluster_lists[index_to_cluster[max_key[0]]] + cluster_lists[index_to_cluster[max_key[1]]]
            
            self._linking_history.insert(0, new_cluster_name)
            
            self._sizes_of_clusters[new_cluster_name] = len1 + len2 

            del cluster_lists[index_to_cluster[max_key[0]]]
            del cluster_lists[index_to_cluster[max_key[1]]]
            cluster_lists[new_cluster_name] = new_cluster

            for x in new_cluster:
                index_to_cluster[x] = new_cluster_name
            
            number_of_iterations += 1
            if number_of_iterations % 100 == 0:
                print "Single link. Iteration #"+str(number_of_iterations)
        
        # Building cluster_lists from index_to_cluster
        if self._initial_cluster_lists != None:
            return self.translate_to_clusters(cluster_lists), None, number_of_iterations
        else:
            return self.translate_to_words(cluster_lists), None, number_of_iterations

    def translate_to_words(self, cluster_lists):
        new_cluster_lists = dict()
        reverse_dict = self._sequencer.get_reversed_dictionary()
        for cluster_name in cluster_lists.keys():
            new_cluster = [reverse_dict[x] for x in cluster_lists[cluster_name]]
            new_cluster_name = reverse_dict[int(cluster_name)]
            new_cluster_lists[new_cluster_name] = new_cluster
        return new_cluster_lists
    
    # Translates this cluster_lists which is a clustering of the initial clusters to a clustering
    # of the words. It does so by uniting the clusters which are found in the same cluster in cluster_lists.
    def translate_to_clusters(self, cluster_lists):
        cluster_lists = self.translate_to_words(cluster_lists)
        new_cluster_lists = dict()
        for cluster_name in cluster_lists.keys():
            cur_cluster = []
            for member in cluster_lists[cluster_name]:
                cur_cluster.extend(self._initial_cluster_lists[member])
            new_cluster_lists[cluster_name] = cur_cluster
        return new_cluster_lists
    
    # Returns the most frequent word in the category if there is a frequency dictionary inside.
    # Otherwise, it returns the first
    def get_representative(self, cluster1_index, cluster2_index):
        if USE_RANDOM_REPRESENTATIVE == True:
            return random.sample([cluster1_index, cluster2_index], 1)[0]
        
        if USE_LAST_MODIFIED_REPRESENTATIVE == True:
            place1 = 1000000
            place2 = 1000000
            if cluster1_index in self._linking_history:
                place1 = self._linking_history.index(cluster1_index)
            if cluster2_index in self._linking_history:
                place2 = self._linking_history.index(cluster2_index)
            if place1 < place2:
                return cluster1_index
            elif place2 < place1:
                return cluster2_index
            else:
                word1 = self._sequencer.inv_get(cluster1_index)
                word2 = self._sequencer.inv_get(cluster2_index)
                if hash(word1) < hash(word2):
                    return cluster1_index
                else:
                    return cluster2_index
        
        if USE_BIGGER_CLUSTER_REPRESENTATIVE == True:
            if self._sizes_of_clusters[cluster1_index] >= self._sizes_of_clusters[cluster2_index]:
                return cluster1_index
            else:
                return cluster2_index
            
        return cluster1_index
# END OF CLASS

# Calculates the cosine metric between two vectors.
# if the flag of the morphology is turned on, it also takes morphological similarity into
# account.
def cos(v1, v2):
    dot_product = 0
    for key in v1.keys():
        dot_product += v1.get(key,0)*v2.get(key, 0)

    norm1 = 0
    for key in v1.keys():
        norm1 += v1[key]*v1[key]

    norm2 = 0
    for key in v2.keys():
        norm2 += v2[key]*v2[key]

    dot_product = dot_product/sqrt(norm1*norm2)
    return dot_product
# END OF METHOD

# Computes the average cosine between the two clusters.
def cos_clusters(cluster1, cluster2, vect_dict):
    average = 0.0
    for v1 in cluster1:
        for v2 in cluster2:
            average += cos(vect_dict[v1], vect_dict[v2])
    return average/(len(cluster1)*len(cluster2))
# END OF METHOD
