#    This is a part of the "Prototype Tagger", an unsupervised POS tagger developed
#    in the Hebrew University by Omri Abend, Roi Reichart and Ari Rappoport.
#    Copyright (C) 2010  Omri Abend, Roi Reichart and Ari Rappoport
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    If you have any questions, feel free to contact us in:
#    omria01@cs.huji.ac.il, roiri@cs.huji.ac.il
#
#    Please read the README file for more details.
#
#!/usr/local/bin/python
import pickle, utils, sys, parameters_dast, re

MINIMAL_MORPHO_CLUSTER_SIZE = 5
MIN_SIGNATURE_SIZE = 2
ENCODE_NULL_SIGNATURE = False

# Creates a morphological clustering.
class morpho_clustering_inducer:
    
    def __init__(self, settings):
        self._settings = settings

    # If the lists are set to null, it reads them from the files specified in the settings.
    def run(self, morfessor_filename = None, uppercase_list = None, non_alphanum_list = None):
        self._type_signature = dict()   # Maps each word to its signature. 
                                        # That is, maps each stem to the set of suffixes
                                        # that this stem ends with.
        self._word_to_stem = dict()
        self._word_to_inclination = dict()
        
        # Loading the uppercase file
        if uppercase_list == None:
            uppercase_list = utils.read_list_from_file(self._settings.get_string("UPPERCASE_LIST_FILENAME"))
        if non_alphanum_list == None:
            non_alphanum_list = utils.read_list_from_file(self._settings.get_string("NON_ALPHANUM_FILENAME"))
        
        if morfessor_filename == None:
            self.create_initial_dicts(self._settings.get_string("MORFESSOR_FILENAME"))
        else:
            self.create_initial_dicts(morfessor_filename)
        
        # Adding the non-alphanumeric signatures
        self.signature_for_non_alpha_numeric(non_alphanum_list)
        
        # Extracting the morpho clusters
        stems_cluster_lists = dict()
        for stem in self._type_signature.keys():
            cur_signature = self._type_signature[stem]
            signature_length = len(cur_signature)
            if signature_length >= MIN_SIGNATURE_SIZE:
                cur_signature.sort()
                temp = stems_cluster_lists.get(str(cur_signature), [])
                temp.append(stem)
                stems_cluster_lists[str(cur_signature)] = temp
        
        utils.remove_small_clusters(stems_cluster_lists, MINIMAL_MORPHO_CLUSTER_SIZE)
        stem_to_signature = utils.get_reversed_dictionary(stems_cluster_lists)
        
        # Now the stems were assigned. We are moving to clustering the rest.
        cluster_lists = dict()
        mapped_stems = []
        for cluster in stems_cluster_lists.values():
            for item in cluster:
                mapped_stems.append(item)
        
        for word in self._word_to_stem.keys():
            cur_stem = self._word_to_stem[word] 
            if cur_stem in mapped_stems:
                if (ENCODE_NULL_SIGNATURE == True or 
                   (ENCODE_NULL_SIGNATURE == False and self._word_to_inclination[word] != "") or
                   ("NON_ALPHANUMERIC" in self._type_signature[word]) ):
                    
                    cur_signature = stem_to_signature[cur_stem]
                    cluster_name = cur_signature + self._word_to_inclination[word]
                    if word in uppercase_list:
                        continue
                    temp = cluster_lists.get(cluster_name, [])
                    temp.append("\""+word+"\"")
                    cluster_lists[cluster_name] = temp
        
        assigned_list = []
        for cluster in cluster_lists.values():
            assigned_list.extend(cluster) 
        
        if self._settings.get_int("INCLUDE_CAPITALIZATION_INFO") == 1:
            uppercase_cluster = []
            for word in uppercase_list:
                if not "\""+word+"\"" in assigned_list:
                    uppercase_cluster.append("\""+word+"\"") 
            cluster_lists["NULL_INCLINATION_UPPER_CASE"] =  uppercase_cluster
                          #"UPPER_CASE_WORDS"] = null_inclination_case_list
        
        return cluster_lists

    def create_initial_dicts(self, morfessor_filename):
        morfessor_file = open(morfessor_filename)
        delim = re.compile("[+\\s]+")
        slash_delim = re.compile("/")
        
        for line in morfessor_file:
            if line.strip()[0] == "#":
                continue
            
            fields = delim.split(line.strip())
            fields = fields[1:]
        
            num_of_stems = 0
            for field in fields:
                if "/STM" in field:
                    num_of_stems += 1
            # if there is more than one stem, we don't use it for the mean time
            if num_of_stems != 1:
                continue
                
            cur_inclination = ""
            cur_stem = None
            cur_word = ""
            for field in fields:
                word_label = slash_delim.split(field)
                if len(word_label) != 2:
                    print "No slash in", line, field
                    continue
                cur_word = cur_word + word_label[0]
                if word_label[1] != "STM":
                    cur_inclination = cur_inclination + "+" + field
                else:
                    cur_stem = word_label[0]
            
            self._word_to_stem[cur_word.lower()] = cur_stem.lower()
            self._word_to_inclination[cur_word.lower()] = cur_inclination
            
            cur_stem = cur_stem.lower()
            cur_entry = self._type_signature.get(cur_stem, [])
            if not cur_inclination in cur_entry:
                cur_entry.append(cur_inclination)
            self._type_signature[cur_stem] = cur_entry
        
        morfessor_file.close()
    # END OF METHOD
    
    # Returns a signature for the non-alphanumeric symbols
    def signature_for_non_alpha_numeric(self, non_alphanumeric_list):
        alphanumeric = re.compile("[0-9,A-Za-z]+")
        for word in non_alphanumeric_list:
            signature = alphanumeric.split(word)
            signature2 = ["NON_ALPHANUMERIC"]
            for field in signature:
                if field != "":
                    signature2.append(field)
            self._type_signature[word] = signature2
            self._word_to_stem[word] = word
            self._word_to_inclination[word] = ""
    # END OF METHOD

# END OF CLASS

# Returns a list of unique values of l
def uniq_list_of_lists(l):
    uniq_l = []
    for item in l:
        item.sort()
        if not item in uniq_l:
            uniq_l.append(item)
    return uniq_l
# END OF METHOD
